/*
 * TimeStamperView.java
 */

package jdbernard.timestamper;

import java.awt.Font;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.ButtonGroup;
import javax.swing.JFileChooser;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.jdesktop.application.Action;
import org.jdesktop.application.ResourceMap;
import org.jdesktop.application.SingleFrameApplication;
import org.jdesktop.application.FrameView;
import sun.security.jca.GetInstance;

/**
 * The application's main frame.
 */
public class TimeStamperView extends FrameView implements MouseMotionListener,
ChangeListener {

    public TimeStamperView(SingleFrameApplication app) {
        super(app);

        initComponents();

        ResourceMap resourceMap = getResourceMap();

        getFrame().setUndecorated(true);

        getFrame().setIconImage(resourceMap.getImageIcon("Application.icon")
                .getImage());

        fileChooser = new JFileChooser();
        notesDialog = new NotesDialog(this, false);
        aboutDialog = new AboutDialog(this.getFrame(), false);

        updateTimer = new Timer();
        updateTimer.scheduleAtFixedRate(new TimerTask() {

            @Override
            public void run() {
                Date currentTime = new Date();
                currentTimeLabel.setText(Timeline.shortFormat.format(currentTime));
                if (mostRecentTask != null) {
                    long seconds = currentTime.getTime() - mostRecentTask.getTime();
                    seconds /= 1000;
                    long minutes = seconds / 60;
                    seconds = seconds % 60;
                    long hours = minutes / 60;
                    minutes %= 60;
                    long days = hours / 24;
                    hours %= 24;
                    StringBuilder sb = new StringBuilder();

                    if (days > 0) sb.append(Long.toString(days) + "day ");
                    if (hours > 0) sb.append(Long.toString(hours) + "hr ");
                    if (minutes > 0) sb.append(Long.toString(minutes) + "min ");
                    sb.append(Long.toString(seconds) + "sec");

                    totalTimeNow.setText(sb.toString());
                } else totalTimeNow.setText("");
            }
        }, 0, 1000);

        saveTimelineTimer = new Timer();
        saveTimelineTimer.scheduleAtFixedRate(new TimerTask() {

            @Override
            public void run() {
                saveTimeline();
            }
        }, 0, 1000*60*30);
        thinTaskFont = taskTextField.getFont().deriveFont(Font.PLAIN);
        boldTaskFont = thinTaskFont.deriveFont(Font.BOLD);

        // refresh new timeline
        refreshDialog();
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        mainPanel = new javax.swing.JPanel();
        currentTaskLabel = new javax.swing.JLabel();
        exitButton = new javax.swing.JButton();
        taskTextField = new javax.swing.JTextField();
        startTimeLabel = new javax.swing.JLabel();
        totalTimeNow = new javax.swing.JLabel();
        currentTimeLabel = new javax.swing.JLabel();
        optionsButton = new javax.swing.JButton();
        notesButton = new javax.swing.JToggleButton();
        optionsMenu = new javax.swing.JPopupMenu();
        saveTimelineMenuItem = new javax.swing.JMenuItem();
        saveTimelineAsMenuItem = new javax.swing.JMenuItem();
        loadTimelineMenuItem = new javax.swing.JMenuItem();
        seperator1 = new javax.swing.JSeparator();
        showNotesDialogCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        showPunchcardCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        seperator2 = new javax.swing.JSeparator();
        showAboutCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();

        mainPanel.setBorder(javax.swing.BorderFactory.createMatteBorder(2, 2, 2, 2, new java.awt.Color(0, 0, 0)));
        mainPanel.setName("mainPanel"); // NOI18N
        mainPanel.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mousePressed(java.awt.event.MouseEvent evt) {
                mainPanelMousePressed(evt);
            }
        });
        mainPanel.addMouseMotionListener(this);

        currentTaskLabel.setFont(currentTaskLabel.getFont().deriveFont(currentTaskLabel.getFont().getStyle() | java.awt.Font.BOLD, currentTaskLabel.getFont().getSize()+2));
        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance(jdbernard.timestamper.TimeStamperApp.class).getContext().getResourceMap(TimeStamperView.class);
        currentTaskLabel.setText(resourceMap.getString("currentTaskLabel.text")); // NOI18N
        currentTaskLabel.setName("currentTaskLabel"); // NOI18N

        javax.swing.ActionMap actionMap = org.jdesktop.application.Application.getInstance(jdbernard.timestamper.TimeStamperApp.class).getContext().getActionMap(TimeStamperView.class, this);
        exitButton.setAction(actionMap.get("quit")); // NOI18N
        exitButton.setBorder(null);
        exitButton.setContentAreaFilled(false);
        exitButton.setFocusPainted(false);
        exitButton.setHideActionText(true);
        exitButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        exitButton.setName("exitButton"); // NOI18N
        exitButton.setRolloverIcon(resourceMap.getIcon("exitButton.rolloverIcon")); // NOI18N

        taskTextField.setFont(taskTextField.getFont().deriveFont(taskTextField.getFont().getStyle() | java.awt.Font.BOLD));
        taskTextField.setText(resourceMap.getString("taskTextField.text")); // NOI18N
        taskTextField.setName("taskTextField"); // NOI18N
        taskTextField.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                taskTextFieldKeyReleased(evt);
            }
        });

        startTimeLabel.setFont(startTimeLabel.getFont().deriveFont(startTimeLabel.getFont().getStyle() | java.awt.Font.BOLD, startTimeLabel.getFont().getSize()+4));
        startTimeLabel.setForeground(resourceMap.getColor("startTimeLabel.foreground")); // NOI18N
        startTimeLabel.setText(resourceMap.getString("startTimeLabel.text")); // NOI18N
        startTimeLabel.setName("startTimeLabel"); // NOI18N

        totalTimeNow.setFont(totalTimeNow.getFont().deriveFont(totalTimeNow.getFont().getStyle() | java.awt.Font.BOLD, totalTimeNow.getFont().getSize()+3));
        totalTimeNow.setForeground(resourceMap.getColor("totalTimeNow.foreground")); // NOI18N
        totalTimeNow.setText(resourceMap.getString("totalTimeNow.text")); // NOI18N
        totalTimeNow.setName("totalTimeNow"); // NOI18N

        currentTimeLabel.setFont(currentTimeLabel.getFont().deriveFont(currentTimeLabel.getFont().getStyle() | java.awt.Font.BOLD, currentTimeLabel.getFont().getSize()+3));
        currentTimeLabel.setForeground(resourceMap.getColor("currentTimeLabel.foreground")); // NOI18N
        currentTimeLabel.setText(resourceMap.getString("currentTimeLabel.text")); // NOI18N
        currentTimeLabel.setName("currentTimeLabel"); // NOI18N

        optionsButton.setIcon(resourceMap.getIcon("optionsButton.icon")); // NOI18N
        optionsButton.setBorder(null);
        optionsButton.setBorderPainted(false);
        optionsButton.setContentAreaFilled(false);
        optionsButton.setFocusPainted(false);
        optionsButton.setHideActionText(true);
        optionsButton.setIconTextGap(0);
        optionsButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        optionsButton.setName("optionsButton"); // NOI18N
        optionsButton.setRolloverIcon(resourceMap.getIcon("optionsButton.rolloverIcon")); // NOI18N
        optionsButton.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseClicked(java.awt.event.MouseEvent evt) {
                optionsButtonMouseClicked(evt);
            }
        });

        notesButton.setAction(actionMap.get("editNotes")); // NOI18N
        notesButton.setHideActionText(true);
        notesButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        notesButton.setName("notesButton"); // NOI18N

        javax.swing.GroupLayout mainPanelLayout = new javax.swing.GroupLayout(mainPanel);
        mainPanel.setLayout(mainPanelLayout);
        mainPanelLayout.setHorizontalGroup(
            mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(mainPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(mainPanelLayout.createSequentialGroup()
                        .addComponent(totalTimeNow)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 58, Short.MAX_VALUE)
                        .addComponent(currentTimeLabel))
                    .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, mainPanelLayout.createSequentialGroup()
                        .addComponent(currentTaskLabel)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(startTimeLabel, javax.swing.GroupLayout.DEFAULT_SIZE, 116, Short.MAX_VALUE))
                    .addComponent(taskTextField, javax.swing.GroupLayout.DEFAULT_SIZE, 275, Short.MAX_VALUE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                    .addComponent(optionsButton)
                    .addComponent(notesButton))
                .addGap(0, 0, 0)
                .addComponent(exitButton))
        );
        mainPanelLayout.setVerticalGroup(
            mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(mainPanelLayout.createSequentialGroup()
                .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(exitButton, javax.swing.GroupLayout.PREFERRED_SIZE, 16, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(optionsButton)
                    .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                        .addComponent(currentTaskLabel)
                        .addComponent(startTimeLabel)))
                .addGap(0, 0, 0)
                .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(mainPanelLayout.createSequentialGroup()
                        .addComponent(taskTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addGap(0, 0, 0)
                        .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(totalTimeNow)
                            .addComponent(currentTimeLabel)))
                    .addComponent(notesButton))
                .addGap(312, 312, 312))
        );

        optionsMenu.setName("optionsMenu"); // NOI18N

        saveTimelineMenuItem.setAction(actionMap.get("saveTimeline")); // NOI18N
        saveTimelineMenuItem.setName("saveTimelineMenuItem"); // NOI18N
        optionsMenu.add(saveTimelineMenuItem);

        saveTimelineAsMenuItem.setAction(actionMap.get("saveTimelineAs")); // NOI18N
        saveTimelineAsMenuItem.setName("saveTimelineAsMenuItem"); // NOI18N
        optionsMenu.add(saveTimelineAsMenuItem);

        loadTimelineMenuItem.setAction(actionMap.get("loadTimeline")); // NOI18N
        loadTimelineMenuItem.setName("loadTimelineMenuItem"); // NOI18N
        optionsMenu.add(loadTimelineMenuItem);

        seperator1.setName("seperator1"); // NOI18N
        optionsMenu.add(seperator1);

        showNotesDialogCheckBoxMenuItem.setAction(actionMap.get("editNotes")); // NOI18N
        showNotesDialogCheckBoxMenuItem.setName("showNotesDialogCheckBoxMenuItem"); // NOI18N
        optionsMenu.add(showNotesDialogCheckBoxMenuItem);

        showPunchcardCheckBoxMenuItem.setAction(actionMap.get("showPunchcard")); // NOI18N
        showPunchcardCheckBoxMenuItem.setName("showPunchcardCheckBoxMenuItem"); // NOI18N
        optionsMenu.add(showPunchcardCheckBoxMenuItem);

        seperator2.setName("seperator2"); // NOI18N
        optionsMenu.add(seperator2);

        showAboutCheckBoxMenuItem.setAction(actionMap.get("showAboutDialog")); // NOI18N
        showAboutCheckBoxMenuItem.setName("showAboutCheckBoxMenuItem"); // NOI18N
        optionsMenu.add(showAboutCheckBoxMenuItem);

        setComponent(mainPanel);
    }// </editor-fold>//GEN-END:initComponents

    private void taskTextFieldKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_taskTextFieldKeyReleased
        if (evt.getKeyCode() == KeyEvent.VK_ENTER) {
            Timeline t = ((TimeStamperApp) getApplication()).getActiveTimeline();
            Date d = new Date();
            t.addMarker(d, taskTextField.getText(), "No comments.");
            startTimeLabel.setText(Timeline.shortFormat.format(d));
            taskTextField.setFont(boldTaskFont);
            mostRecentTask = d;
            fireChangeEvent();
        } else {
            taskTextField.setFont(thinTaskFont);
        }
    }//GEN-LAST:event_taskTextFieldKeyReleased

    private void mainPanelMousePressed(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_mainPanelMousePressed
        mousePressRelativeToFrame = evt.getPoint();
    }//GEN-LAST:event_mainPanelMousePressed

    private void optionsButtonMouseClicked(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_optionsButtonMouseClicked
        optionsMenu.show(evt.getComponent(), evt.getX(), evt.getY());
    }//GEN-LAST:event_optionsButtonMouseClicked

    @Action
    public void editNotes() {
        notesVisible = !notesVisible;
        notesDialog.setVisible(notesVisible);
        notesButton.setSelected(notesVisible);
        showNotesDialogCheckBoxMenuItem.setSelected(notesVisible);
    }

    @Action
    public void showPunchcard() {
        punchcardVisible = !punchcardVisible;
        showPunchcardCheckBoxMenuItem.setSelected(punchcardVisible);
        getPunchcardDisplayDialog().setVisible(punchcardVisible);
    }

    @Action
    public void showAboutDialog() {
        aboutDialogVisible = !aboutDialogVisible;
        showAboutCheckBoxMenuItem.setSelected(aboutDialogVisible);
        aboutDialog.setVisible(aboutDialogVisible);
    }

    public void setNotesForActiveTask(String notes) {
        Timeline t = ((TimeStamperApp) getApplication()).getActiveTimeline();
        Timeline.TimelineMarker tm = t.getLastMarker(new Date());
        if (tm == null) return;
        tm.setNotes(notes);
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel currentTaskLabel;
    private javax.swing.JLabel currentTimeLabel;
    private javax.swing.JButton exitButton;
    private javax.swing.JMenuItem loadTimelineMenuItem;
    private javax.swing.JPanel mainPanel;
    private javax.swing.JToggleButton notesButton;
    private javax.swing.JButton optionsButton;
    private javax.swing.JPopupMenu optionsMenu;
    private javax.swing.JMenuItem saveTimelineAsMenuItem;
    private javax.swing.JMenuItem saveTimelineMenuItem;
    private javax.swing.JSeparator seperator1;
    private javax.swing.JSeparator seperator2;
    private javax.swing.JCheckBoxMenuItem showAboutCheckBoxMenuItem;
    private javax.swing.JCheckBoxMenuItem showNotesDialogCheckBoxMenuItem;
    private javax.swing.JCheckBoxMenuItem showPunchcardCheckBoxMenuItem;
    private javax.swing.JLabel startTimeLabel;
    private javax.swing.JTextField taskTextField;
    private javax.swing.JLabel totalTimeNow;
    // End of variables declaration//GEN-END:variables

    private Point mousePressRelativeToFrame;
    private JFileChooser fileChooser;
    private Timer updateTimer;
    private Timer saveTimelineTimer;
    private Font boldTaskFont;
    private Font thinTaskFont;
    private Date mostRecentTask;
    private NotesDialog notesDialog;
    private PunchcardDisplayDialog punchcardDisplayDialog;
    private AboutDialog aboutDialog;
    private boolean notesVisible = false;
    private boolean punchcardVisible = false;
    private boolean aboutDialogVisible = false;
    private ArrayList<ChangeListener> changeListeners =
            new ArrayList<ChangeListener>();

    private PunchcardDisplayDialog getPunchcardDisplayDialog() {
        if (punchcardDisplayDialog == null) {
            punchcardDisplayDialog = new PunchcardDisplayDialog(this.getFrame(),
                    false, Calendar.getInstance());
            punchcardDisplayDialog.addChangeListener(this);
            addChangeListener(punchcardDisplayDialog);
        }
        return punchcardDisplayDialog;
    }

    public void mouseDragged(MouseEvent e) {
        getFrame().setLocation(calculateWindowMovement(e.getLocationOnScreen(),
                mousePressRelativeToFrame, getFrame().getBounds(),
                new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
    }

    public static Point calculateWindowMovement(Point currentMousePoint,
    Point mousePressRelativeToWindow, Rectangle windowBounds,
    Rectangle... snapBoxes) {

        // this is the point we will compute as the new upper left
        // coordinate of the frame. It needs to be translated some to account
        // for the fact that the user can press the mouse anywhere on the
        // main panel.
        Point currentWindowPoint = (Point) currentMousePoint.clone();
        // find out where the new point should be, ignoreing screen bounds
        currentWindowPoint.translate(-mousePressRelativeToWindow.x,
            -mousePressRelativeToWindow.y);

        Point finalWindowPoint = (Point) currentWindowPoint.clone();

        // snap tests. In the event of a conflict in snaps positions, the
        // last snap wins
        int wUp = windowBounds.y;
        int wDown = windowBounds.y + windowBounds.height;
        int wLeft = windowBounds.x;
        int wRight = windowBounds.x + windowBounds.width;

//        currentTaskLabel.setText("U:" + wUp + " D:" + wDown + " L:" + wLeft
//                + " R:" + wRight);

        for (Rectangle r : snapBoxes) {
            int rUp = r.y;
            int rDown = r.y + r.height;
            int rLeft = r.x;
            int rRight = r.x + r.width;

            // check to see if the window is near any of the snap boundaries
            // if it is, 'snap' the final point to that boundary.
            if (Math.abs(rUp - wUp) < 20)   // top of window to top of rect
                finalWindowPoint.y = rUp;
            else if (Math.abs(rUp - wDown) < 20) // bot of w to top of r
                finalWindowPoint.y = rUp - windowBounds.height;
            else if (Math.abs(rDown - wUp) < 20) // top of w to bot of r
                finalWindowPoint.y = rDown;
            else if (Math.abs(rDown - wDown) < 20)   // bot of w to bot of r
                finalWindowPoint.y = rDown - windowBounds.height;

            if (Math.abs(rLeft - wLeft) < 20)   // left of w to left of r
                finalWindowPoint.x = rLeft;
            else if (Math.abs(rLeft - wRight) < 20)  // right of w to left of r
                finalWindowPoint.x = rLeft - windowBounds.width;
            else if (Math.abs(rRight - wLeft) < 20)  // left of w to right of r
                finalWindowPoint.x = rRight;
            else if (Math.abs(rRight - wRight) < 20) // right of w to right of r
                finalWindowPoint.x = rRight - windowBounds.width;
        }

        // this point represents a backwards version of the initial calculation
        // to find the finalPoint. It says, based on the current final point,
        // assume I moved the frame to that point. Where, relative to that point
        // should the mouse be, based on how far it was relative to the point
        // when the user pressed it.
        Point whereMouseShouldBe = (Point) finalWindowPoint.clone();
        whereMouseShouldBe.translate(mousePressRelativeToWindow.x,
                mousePressRelativeToWindow.y);

        // if the actual mouse location is different from the expected location,
        // then we know that the snapping behaviour has altered the frames new
        // placement. If this alteration is too large (30 px apart in this case)
        // then we know the user is trying to pull the frame out of its snapped
        // position. In this case, we want to ignore the snap calculations and
        // base the new frame location entirely on the current mouse location
        if (whereMouseShouldBe.distance(currentMousePoint) > 30) {
            finalWindowPoint = (Point) currentMousePoint.clone();
            finalWindowPoint.translate(-mousePressRelativeToWindow.x,
                    -mousePressRelativeToWindow.y);
        }

        return finalWindowPoint;
    }

    public void mouseMoved(MouseEvent e) {
    }

    public void stateChanged(ChangeEvent ce) {
        refreshDialog();
    }

    public void addChangeListener(ChangeListener cl) {
        changeListeners.add(cl);
    }

    public boolean removeChangeListener(ChangeListener cl) {
        return changeListeners.remove(cl);
    }

    private void fireChangeEvent() {
        ChangeEvent ce = new ChangeEvent(this);
        for (ChangeListener cl : changeListeners)
            cl.stateChanged(ce);

        refreshDialog();
    }

    @Action
    public void saveTimeline() {
        ((TimeStamperApp) getApplication()).saveTimeline();
    }

    @Action
    public void saveTimelineAs() {
        if (fileChooser.showSaveDialog(getFrame())!=JFileChooser.APPROVE_OPTION)
            return;

        ((TimeStamperApp) getApplication()).saveTimelineToFile(
                fileChooser.getSelectedFile().getAbsolutePath());
    }

    @Action
    public void loadTimeline() {
        if (fileChooser.showOpenDialog(getFrame())!=JFileChooser.APPROVE_OPTION)
            return;

        ((TimeStamperApp) getApplication()).loadTimeline(
                fileChooser.getSelectedFile().getAbsolutePath());

        fireChangeEvent();
    }

    public void refreshDialog() {
        Timeline t = ((TimeStamperApp) getApplication()).getActiveTimeline();
        Timeline.TimelineMarker lastMarker = t.getLastMarker(new Date());

        if (lastMarker != null) {
            mostRecentTask = lastMarker.getTimestamp();
            startTimeLabel.setText(Timeline.shortFormat.format(lastMarker.getTimestamp()));

            taskTextField.setText(lastMarker.getMark());
            notesDialog.setNotes(lastMarker.getNotes());
        }
    }
}
